/*****************************************************************************
 *
 * FILE:	wkf.c
 * DESCRIPTION:	WALL's Kanji Filter
 * DATE:	Sat, Jan 15 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.com
 * URL:		http://www.MysticWALL.com/cover.html
 * COPYRIGHT:	(C) 1997-2000 졿Kouichi ABE (WALL), All rights reserved.
 * $Id: wkf.c,v 1.1.2.2 2001/01/12 07:58:43 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */
#include <stdio.h>
#include <stdlib.h>
#ifndef _WIN32
#include <unistd.h>
#endif	/* _WIN32 */
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#if defined(_WIN32) || defined(__EMX__)
#include <sys/utime.h>
#include <fcntl.h>
#else
#include <utime.h>
#endif	/* _WIN32 __EMX__ */
#include <signal.h>

#define	MAIN_MODULE
#include "wkf.h"
#include "version.h"
#undef	MAIN_MODULE


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	FILE_LEN	256	/* max length of filename	*/
#define	EOL		'\0'	/* End of Line */

typedef char *	STR;

typedef enum {	/* error code type */
  ERR_NO_CONVERT_OPTION,	/* no option to convert kanji code */
  ERR_MULTIPLE_CONVERT_OPTION,	/* multiple option to convert */
  ERR_MULTIPLE_INPUT_CODE_OPTION, /* multiple option of input kanji code */
  ERR_MULTIPLE_LINE_END_OPTION	/* multiple option of the code of line end */
} error_t;


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
const char	Program[] = "wkf";


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	savePermission(STR, STR);
static void	saveTimeStamp(STR, STR);
static STR	tmpName(STR);
static STR	backupName(STR);
static void	printErrorMessage(error_t);
static void	copyright(void);
static void	usage(void);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	savePermission
 * DESCRIPTION:		save permission of file
 * ARGUMENTS:		input filename, output filename
 * RETURN VALUE:	none
 */
static void
savePermission(org_name, new_name)
	STR	org_name;
	STR	new_name;
{
  struct stat		sb;

  if (stat(org_name, &sb) != 0) {
    fprintf(stderr, "Warning: "
	"%s couldn't set permission on file %s\n", Program, new_name);
    return;
  }

  if (chmod(new_name, sb.st_mode) != 0) {
    fprintf(stderr, "Warniing: "
	"%s couldn't set permission on file %s\n", Program, new_name);
    return;
  }
}

/*
 * FUNCTION NAME:	saveTimeStamp
 * DESCRIPTION:		save time-stamp of file
 * ARGUMENTS:		input filename, output filename
 * RETURN VALUE:	none
 */
static void
saveTimeStamp(org_name, new_name)
	STR	org_name;
	STR	new_name;
{
  struct stat		sb;
  struct utimbuf	tb;

  if (stat(org_name, &sb) != 0) {
    fprintf(stderr, "Warning: "
	"%s couldn't set time-stamp on file %s\n", Program, new_name);
    return;
  }

  tb.actime  = sb.st_atime;
  tb.modtime = sb.st_mtime;

  if (utime(new_name, &tb) != 0) {
    fprintf(stderr, "Warning: "
	"%s couldn't set time-stamp on file %s\n", Program, new_name);
    return;
  }
}

/*
 * FUNCTION NAME:	tmpName
 * DESCRIPTION:		make a temporary filename
 * ARGUMENTS:		filename
 * RETURN VALUE:	temporary filename
 */
static STR
tmpName(fname)
	STR	fname;
{
  static char	buf[FILE_LEN];

  memset((STR)buf, (int)0, (size_t)FILE_LEN);
  sprintf(buf, "%s.wkf~", fname);

  return buf;
}

/*
 * FUNCTION NAME:	backupName
 * DESCRIPTION:		make a backup filename
 * ARGUMENTS:		filename
 * RETURN VALUE:	backup filename
 */
static STR
backupName(fname)
	STR	fname;
{
  static char	buf[FILE_LEN];

  memset((STR)buf, (int)0, (size_t)FILE_LEN);
  sprintf(buf, "%s.bak", fname);

  return buf;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	printErrorMessage
 * DESCRIPTION:		print error message
 * ARGUMENTS:		error type
 * RETURN VALUE:	none
 */
static void
printErrorMessage(err)
	error_t	err;
{
  switch (err) {
    case ERR_NO_CONVERT_OPTION:
      fprintf(stderr,
		"%s: no convert option -- needs 'j' or 'e' or 's' option.\n",
		Program);
      break;
    case ERR_MULTIPLE_CONVERT_OPTION:
      fprintf(stderr,
		"%s: bad option -- j, e, s, c, C selected only one.\n",
		Program);
      break;
    case ERR_MULTIPLE_INPUT_CODE_OPTION:
      fprintf(stderr,
		"%s: bad option -- J, E, S selected only one.\n",
		Program);
      break;
    case ERR_MULTIPLE_LINE_END_OPTION:
      fprintf(stderr,
		"%s: bad option -- M, D, U selected only one.\n",
		Program);
      break;
  }
}

/*
 * FUNCTION NAME:	copyright
 * DESCRIPTION:		print copyright
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
copyright(void)
{
  fprintf(stderr,
	  "%s\n"
	  "Version %s\n"
	  "Copyright (C) 1997-2001 Kouichi ABE (WALL)\n"
	  "%s\n",
	  SOFTWARE_NAME,
	  SOFTWARE_VERSION,
	  SOFTWARE_ADDRESS);
}

/*
 * FUNCTION NAME:	usage
 * DESCRIPTION:		print usage
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
usage(void)
{
  fprintf(stderr,
	  "usage: "
	  "%s [-{j|e|s|c|C}] [-zAT] [-{J|E|S}] [-{M|D|U}] [<file>...]\n",
	  Program);
  exit(0);
}


/*****************************************************************************
 *****************************************************************************/

/*
 *	MAIN FUNCTION 
 */
int
main(argc, argv)
	int	argc;
	char	**argv;
{
  int		i;
  STR		tmp;		/* temporary filename */
  STR		bck;		/* backup filename */
  conv_t	convert;	/* return value of convert function */
  kcode_t	kin;		/* input kanji code */
  kcode_t	kout;		/* output kanji code */
  bool_t	checkcode;	/* flag to check kanji code */
  bool_t	showfname;	/* flag to show filename */
  bool_t	backup;		/* flag to do backup */
  bool_t	savetstamp;	/* flag to save the time-stamp of file */
  bool_t	lineend;	/* type of the code of line end	*/
  FILE *	fin;		/* opened file */
  FILE *	bfout;		/* backup file */
  char		c;

  /* set default value */
  kin		= KC_UNKNOWN;
  kout		= KC_UNKNOWN;
  checkcode	= FALSE;	/* no checking kanji code */
  showfname	= FALSE;	/* no showing filename */
  backup	= FALSE;	/* no backup */
  savetstamp	= FALSE;	/* no saving time-stamp */
  lineend	= FALSE;	/* default use (UNIX type) */

  /* addressing arguments */
  while ((--argc > 0) && ((*++argv)[0] == '-')) {
    while ((c = *++argv[0]) != EOL) {
      switch (c) {
	case 'h':	/* show usage */
	  copyright();
	  usage();
	case 'v':	/* show version */
	  copyright();
	  exit(0);
	case 'C':	/* check kanji code and show filename */
	  showfname = TRUE;
	case 'c':	/* check kanji code */
	  checkcode = TRUE;
	  if (kin != KC_UNKNOWN || kout != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  break;
	case 'j':	/* convert to JIS */
	  if (checkcode) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  else if (kout != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  kout = KC_JIS;
	  break;
	case 'e':	/* convert to Japanese EUC */
	  if (checkcode) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  else if (kout != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  kout = KC_EUC;
	  break;
	case 's':	/* convert to Shift-JIS */
	  if (checkcode) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  else if (kout != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  kout = KC_SJIS;
	  break;
	case 'z':	/* convert 2byte alphabet and figures to 1byte ones */
	  wkfConvertZenkaku2ASCII(TRUE);
	  break;
	case 'A':	/* make a backup file */
	  backup = TRUE;
	  break;
	case 'T':	/* save the time-stamp */
	  savetstamp = TRUE;
	  break;
	case 'M':	/* output the line end with Macintosh format */
	  if (lineend) {
	    printErrorMessage(ERR_MULTIPLE_LINE_END_OPTION);
	    usage();
	  }
	  lineend = TRUE;
	  wkfSetLineEndCode(LE_CR);
	  break;
	case 'D':	/* output the line end with DOS format */
	  if (lineend) {
	    printErrorMessage(ERR_MULTIPLE_LINE_END_OPTION);
	    usage();
	  }
	  lineend = TRUE;
	  wkfSetLineEndCode(LE_CRLF);
	  break;
	case 'U':	/* output the line end with UNIX format */
	  if (lineend) {
	    printErrorMessage(ERR_MULTIPLE_LINE_END_OPTION);
	    usage();
	  }
	  lineend = TRUE;
	  wkfSetLineEndCode(LE_LF);
	  break;
	case 'J':	/* input kanji code is JIS */
	  if (kin != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_INPUT_CODE_OPTION);
	    usage();
	  }
	  kin = KC_JIS;
	  break;
	case 'E':	/* input kanji code is Japanese EUC */
	  if (kin != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_INPUT_CODE_OPTION);
	    usage();
	  }
	  kin = KC_EUC;
	  break;
	case 'S':	/* input kanji code is Shift-JIS */
	  if (kin != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_INPUT_CODE_OPTION);
	    usage();
	  }
	  kin = KC_SJIS;
	  break;
	default:	/* illegal option */
	  fprintf(stderr, "%s: illegal option -- %c\n", Program, c);
	  argc = -1;
	  break;
      }
    }
  }

  /* check option to convert kanji code */
  if (checkcode == FALSE &&	/* no check kanji code */
      kout == KC_UNKNOWN) {	/* no option to convert kanji code */
    printErrorMessage(ERR_NO_CONVERT_OPTION);
    usage();
  }

  /* processing every file */
  if (argc > 0) {
    for (i = 0; i < argc; i++) {
      signal(SIGINT,  SIG_IGN);
      fin = fopen(argv[i], "r");
      if (fin == NULL) {
	fprintf(stderr, "couldn't open %s\n", argv[i]);
	continue;
      }
#if defined(_WIN32) || defined(__EMX__)
	/*  D,M,U Ū˻ꤷʤ CRLF */
	_setmode(fileno(fin), O_BINARY);  /* ϤϾ binary open */
	if (lineend == FALSE) {
	  wkfSetLineEndCode(LE_CRLF);
	 _setmode(fileno(stdout), O_BINARY);
	}
#endif	/* _WIN32 __EMX__ */
      setvbuf(fin, (char*)NULL, _IOFBF, 0);
      if (checkcode) {	/* only check kanji code */
	if (showfname) {
	  fprintf(stdout, "%-16s\t", argv[i]);
	  fflush(stdout);
	}
	wkfPrintGuessedKanjiCodeOfFile(fin, FALSE);
	continue;
      }
      if (backup) {
	tmp   = tmpName(argv[i]);	/* get a temporary filename */
	bck   = backupName(argv[i]);	/* get a backup filename */
	bfout = fopen(tmp, "w");
	if (bfout == NULL) {
	  fprintf(stderr, "couldn't open backup file for %s\n", argv[i]);
	  continue;
	}
#if defined(_WIN32) || defined(__EMX__)
	/*  D,M,U Ū˻ꤷʤ CRLF */
	if (lineend == FALSE) {
	  wkfSetLineEndCode(LE_CRLF);
	  _setmode(fileno(bfout), O_BINARY);
	}
#endif	/* _WIN32 __EMX__ */
	setvbuf(bfout, (char*)NULL, _IOFBF, 0);
	convert = wkfConvertKanjiCodeOfFile(kin, fin, kout, bfout);
	fflush(bfout);
	fclose(bfout);
	fclose(fin);
	if (convert == CONV_OK) {
#ifdef _WIN32          
	  unlink(bck);
#endif          
	  /* ̾ѹΥ顼å򤷤褦	*/
	  rename(argv[i], bck);
	  rename(tmp, argv[i]);
	  if (savetstamp) {		/* save file time-stamp */
	    saveTimeStamp(bck, argv[i]);
	  }
	  savePermission(bck, argv[i]);	/* save file permission */
	}
	else {
	  unlink(tmp);	/* remove the temporary file */
	}
      }
      else {
	setvbuf(stdout, (char*)NULL, _IOFBF, 0);
	convert = wkfConvertKanjiCodeOfFile(kin, fin, kout, stdout);
	fclose(fin);
	fflush(stdout);
	if (convert != CONV_OK) {
	  fprintf(stderr, "%s: no converted ...\n", argv[i]);
	}
      }
      signal(SIGINT,  SIG_DFL);
    }
  }
  else if (argc == 0) {	/* for standard input */
    if (checkcode) {	/* only check kanji code */
      if (showfname) {
	fprintf(stdout, "stdin\t");
	fflush(stdout);
      }
      wkfPrintGuessedKanjiCodeOfFile(stdin, FALSE);
      fflush(stdout);
    }
    else {
#if defined(_WIN32) || defined(__EMX__)
	/*  D,M,U Ū˻ꤷʤ CRLF */
	_setmode(fileno(stdin), O_BINARY);  /* ϤϾ binary open */
	if (lineend == FALSE) {
	  wkfSetLineEndCode(LE_CRLF);
	 _setmode(fileno(stdout), O_BINARY);
	}
#endif	/* _WIN32 __EMX__ */
      setvbuf(stdin, (char*)NULL, _IOFBF, 0);
      setvbuf(stdout, (char*)NULL, _IOFBF, 0);
      wkfConvertKanjiCodeOfFile(kin, stdin, kout, stdout);
      fflush(stdout);
    }
  }
  else {	/* show usage */
    usage();
  }

  return 0;
}
